"""Utilities module."""
from __future__ import annotations

import asyncio
import logging
from collections.abc import Callable, Iterable
from typing import Any, TypeVar
from warnings import warn

_LOGGER = logging.getLogger(__name__)

_T = TypeVar("_T")


def first_or_none(lst: Iterable[_T], predicate: Callable[[_T], Any]) -> _T | None:
    """Return the first occurrence or `None` of an iterable, given a predicate."""
    filter_iter = filter(predicate, lst)
    return next(filter_iter, None)


def add_async_job(target: Callable, *args: Any) -> asyncio.Task | asyncio.Future:
    """Add a callable to the event loop."""
    loop = asyncio.get_event_loop()
    task: asyncio.Future | asyncio.Task

    if asyncio.iscoroutine(target):
        task = loop.create_task(target)
    elif asyncio.iscoroutinefunction(target):
        task = loop.create_task(target(*args))
    else:
        task = loop.run_in_executor(None, target, *args)

    return task


def send_deprecation_warning(old_name: str, new_name: str) -> None:
    """Send a deprecation warning."""
    message = f"{old_name} has been deprecated in favor of {new_name}, the alias will be removed in the future"
    warn(
        message,
        DeprecationWarning,
        stacklevel=2,
    )
    _LOGGER.warning(message)



"""File for various patches to PubNub until fixed in module."""
from pubnub import pubnub_asyncio
from pubnub.endpoints.presence.heartbeat import Heartbeat
from pubnub.enums import PNHeartbeatNotificationOptions, PNStatusCategory
from pubnub.models.consumer.common import PNStatus
from pubnub.pubnub_asyncio import (
    AsyncioSubscriptionManager,
    PubNubAsyncioException,
    asyncio,
)

# pylint: disable=protected-access, fixme


async def patched_perform_heartbeat_loop(self: AsyncioSubscriptionManager) -> None:
    """
    Patched `_perform_heartbeat_loop` function on pubnub's `AsyncioSubscriptionManager` class.

    Correctly adheres to setting for heartbeat_notification_options.

    See https://github.com/pubnub/python/pull/101 for when this can be removed.
    """
    if self._heartbeat_call is not None:
        # TODO: cancel call
        pass

    cancellation_event = asyncio.Event()
    state_payload = self._subscription_state.state_payload()
    presence_channels = self._subscription_state.prepare_channel_list(False)
    presence_groups = self._subscription_state.prepare_channel_group_list(False)

    if len(presence_channels) == 0 and len(presence_groups) == 0:
        return

    try:
        heartbeat_call = (
            Heartbeat(self._pubnub)
            .channels(presence_channels)
            .channel_groups(presence_groups)
            .state(state_payload)
            .cancellation_event(cancellation_event)
            .future()
        )

        envelope = await heartbeat_call

        heartbeat_verbosity = self._pubnub.config.heartbeat_notification_options
        if heartbeat_verbosity == PNHeartbeatNotificationOptions.ALL or (
            envelope.status.is_error()
            and heartbeat_verbosity == PNHeartbeatNotificationOptions.FAILURES
        ):
            self._listener_manager.announce_status(envelope.status)

        if envelope.status is not None and envelope.status.category in [
            PNStatusCategory.PNTimeoutCategory,
            PNStatusCategory.PNNetworkIssuesCategory,
        ]:
            pass

    except PubNubAsyncioException:
        pass

    finally:
        cancellation_event.set()


pubnub_asyncio.AsyncioSubscriptionManager._perform_heartbeat_loop = (
    patched_perform_heartbeat_loop
)


def patched_is_error(self: PNStatus) -> bool:
    """
    Patched `is_error` function on pubnub's `PNStatus` class.

    See https://github.com/pubnub/python/pull/101 for when this can be removed.
    """
    return bool(self.error)


PNStatus.is_error = patched_is_error



"""Module that implements the DoorLock class."""
from __future__ import annotations

from ..const import ZWaveDeviceAttribute as Attributes
from . import BypassTamperDevice, VivintDevice


class DoorLock(BypassTamperDevice, VivintDevice):
    """Represents a vivint door lock device."""

    @property
    def is_locked(self) -> bool:
        """Return True if door lock is locked."""
        return bool(self.data[Attributes.STATE])

    @property
    def node_online(self) -> bool:
        """Return True if the node is online."""
        return bool(self.data[Attributes.ONLINE])

    async def set_state(self, locked: bool) -> None:
        """Set door lock's state."""
        assert self.alarm_panel
        await self.vivintskyapi.set_lock_state(
            self.alarm_panel.id, self.alarm_panel.partition_id, self.id, locked
        )

    async def lock(self) -> None:
        """Lock the door lock."""
        await self.set_state(True)

    async def unlock(self) -> None:
        """Unlock the door lock."""
        await self.set_state(False)



"""Module that implements the VivintPubNubSubscribeListener class."""
from __future__ import annotations

import logging
from collections.abc import Callable
from enum import IntEnum, unique
from typing import Any

from pubnub.callbacks import SubscribeCallback
from pubnub.models.consumer.pubsub import PNMessageResult
from pubnub.pubnub_asyncio import PubNubAsyncio

from .pubnub_patches import PNStatus

PN_SUBSCRIBE_KEY = "sub-c-6fb03d68-6a78-11e2-ae8f-12313f022c90"
PN_CHANNEL = "PlatformChannel"

_LOGGER = logging.getLogger(__name__)


class VivintPubNubSubscribeListener(SubscribeCallback):
    """PubNub Subscribe Listener."""

    def __init__(self, message_received_callback: Callable):
        """Initialize the PubNub subscription."""
        super().__init__()
        self.__message_received_callback = message_received_callback

    def presence(self, pubnub: PubNubAsyncio, presence: Any) -> None:
        """Handle presence update."""
        _LOGGER.debug("Recieved new presence: %s", presence)

    def status(
        self,
        pubnub: PubNubAsyncio,
        status: PNStatus,
    ) -> None:
        """Handle a status update."""
        operation = PubNubOperationType(status.operation).name
        category = PubNubStatusCategory(status.category).name
        if status.is_error():
            _LOGGER.error(
                "PubNub status error - operation: %s, category: %s, error: %s",
                operation,
                category,
                status.error_data.information,
            )
        else:
            _LOGGER.debug(
                "PubNub status update - operation: %s, category: %s",
                operation,
                category,
            )

    def message(
        self,
        pubnub: PubNubAsyncio,
        message: PNMessageResult,
    ) -> None:
        """Handle a message received."""
        self.__message_received_callback(message.message)


@unique
class PubNubStatusCategory(IntEnum):
    """PubNub status category."""

    UNKNOWN = 1
    ACKNOWLEDGMENT = 2
    ACCESS_DENIED = 3
    TIMEOUT = 4
    NETWORK_ISSUES = 5
    CONNECTED = 6
    RECONNECTED = 7
    DISCONNECTED = 8
    UNEXPECTED_DISCONNECT = 9
    CANCELLED = 10
    BAD_REQUEST = 11
    MALFORMED_FILTER_EXPRESSION = 12
    MALFORMED_RESPONSE = 13
    DECRYPTION_ERROR = 14
    TLS_CONNECTION_FAILED = 15
    TLS_UNTRUSTED_CERTIFICATE = 16
    INTERNAL_EXCEPTION = 17

    # Handle unknown/future status category
    UNEXPECTED_STATUS_CATEGORY = -1

    @classmethod
    def _missing_(cls, value: Any) -> PubNubStatusCategory:
        if value is not None:
            _LOGGER.error("Unexpected status category value: %s", value)
        return cls.UNEXPECTED_STATUS_CATEGORY


@unique
class PubNubOperationType(IntEnum):
    """PubNub operation type."""

    SUBSCRIBE = 1
    UNSUBSCRIBE = 2
    PUBLISH = 3
    HISTORY = 4
    WHERE_NOW = 5

    HEARTBEAT = 6
    SET_STATE = 7
    ADD_CHANNELS_TO_GROUP = 8
    REMOVE_CHANNELS_FROM_GROUP = 9
    CHANNEL_GROUPS = 10
    REMOVE_GROUP = 11
    CHANNELS_FOR_GROUP = 12
    PUSH_NOTIFICATION_ENABLED_CHANNELS = 13
    ADD_PUSH_NOTIFICATIONS_ON_CHANNELS = 14
    REMOVE_PUSH_NOTIFICATIONS_FROM_CHANNELS = 15
    REMOVE_ALL_PUSH_NOTIFICATIONS = 16
    TIME = 17

    HERE_NOW = 18
    GET_STATE = 19
    ACCESS_MANAGER_AUDIT = 20
    ACCESS_MANAGER_GRANT = 21
    ACCESS_MANAGER_REVOKE = 22
    HISTORY_DELETE = 23
    MESSAGE_COUNT = 24
    FIRE = 25
    SIGNAL = 26

    ACCESS_MANAGER_REVOKE_TOKEN = 40
    ACCESS_MANAGER_GRANT_TOKEN = 41

    ADD_MESSAGE_ACTION = 42
    GET_MESSAGE_ACTIONS = 43
    DELETE_MESSAGE_ACTION = 44
    FETCH_MESSAGES = 45

    GET_FILES_ACTION = 46
    DELETE_FILE = 47
    GET_FILE_DOWNLOAD_URL_ACTION = 48
    FETCH_FILE_UPLOAD_S3_DATA_ACTION = 49
    DOWNLOAD_FILE_ACTION = 50
    SEND_FILE_ACTION = 51
    SEND_FILE_NOTIFICATION = 52

    SET_UUID_METADATA = 53
    GET_UUID_METADATA = 54
    REMOVE_UUID_METADATA = 55
    GET_ALL_UUID_METADATA = 56

    SET_CHANNEL_METADATA = 57
    GET_CHANNEL_METADATA = 58
    REMOVE_CHANNEL_METADATA = 59
    GET_ALL_CHANNEL_METADATA = 60

    SET_CHANNEL_MEMBERS = 61
    GET_CHANNEL_MEMBERS = 62
    REMOVE_CHANNEL_MEMBERS = 63
    MANAGE_CHANNEL_MEMBERS = 64

    SET_MEMBERSHIPS = 65
    GET_MEMBERSHIPS = 66
    REMOVE_MEMBERSHIPS = 67
    MANAGE_MEMBERSHIPS = 68

    # Handle unknown/future operation type
    UNEXPECTED_OPERATION_TYPE = -1

    @classmethod
    def _missing_(cls, value: Any) -> PubNubOperationType:
        if value is not None:
            _LOGGER.error("Unexpected operation type value: %s", value)
        return cls.UNEXPECTED_OPERATION_TYPE



"""Module that implements the VivintSkyApi class."""
from __future__ import annotations

import json
import logging
import ssl
from collections.abc import Callable
from typing import Any

import aiohttp
import certifi
from aiohttp import ClientResponseError
from aiohttp.client import _RequestContextManager

from .const import (
    AuthenticationResponse,
    MfaVerificationResponse,
    SwitchAttribute,
    VivintDeviceAttribute,
)
from .enums import ArmedState, GarageDoorState, ZoneBypass
from .exceptions import (
    VivintSkyApiAuthenticationError,
    VivintSkyApiError,
    VivintSkyApiMfaRequiredError,
)

_LOGGER = logging.getLogger(__name__)

VIVINT_API_ENDPOINT = "https://www.vivintsky.com/api"
VIVINT_MFA_ENDPOINT = (
    "https://www.vivintsky.com/platform-user-api/v0/platformusers/2fa/validate"
)


class VivintSkyApi:
    """Class to communicate with the VivintSky API."""

    def __init__(
        self,
        username: str,
        password: str,
        persist_session: bool = False,
        client_session: aiohttp.ClientSession | None = None,
    ) -> None:
        """Initialize the VivintSky API."""
        self.__username = username
        self.__password = password
        self.__persist_session = persist_session
        self.__client_session = client_session or self.__get_new_client_session()
        self.__has_custom_client_session = client_session is not None
        self.__mfa_pending = False

    def is_session_valid(self) -> bool:
        """Return the state of the current session."""
        cookies = self.__client_session.cookie_jar.filter_cookies(VIVINT_API_ENDPOINT)
        return cookies.get("s") is not None

    async def connect(self) -> dict:
        """Connect to VivintSky Cloud Service."""
        if self.__has_custom_client_session and self.is_session_valid():
            authuser_data = await self.get_authuser_data()
        else:
            authuser_data = await self.__get_vivintsky_session(
                self.__username, self.__password, self.__persist_session
            )
        if not authuser_data:
            raise VivintSkyApiAuthenticationError("Unable to login to Vivint")
        return authuser_data

    async def disconnect(self) -> None:
        """Disconnect from VivintSky Cloud Service."""
        if not self.__has_custom_client_session:
            await self.__client_session.close()

    async def verify_mfa(self, code: str) -> None:
        """Verify multi-factor authentication code."""
        resp = await self.__post(
            VIVINT_MFA_ENDPOINT,
            data=json.dumps({"code": code}),
        )
        if resp is not None:
            self.__mfa_pending = False

    async def get_authuser_data(self) -> dict:
        """
        Get the authuser data.

        Poll the Vivint authuser API endpoint resource to gather user-related data including enumeration of the systems
        that user has access to.
        """
        resp = await self.__get("authuser")
        if not resp:
            raise VivintSkyApiAuthenticationError("Missing auth user data")
        return resp

    async def get_panel_credentials(self, panel_id: int) -> dict:
        """Get the panel credentials."""
        resp = await self.__get(f"panel-login/{panel_id}")
        if not resp:
            raise VivintSkyApiAuthenticationError(
                "Unable to retrieve panel credentials."
            )
        return resp

    async def get_system_data(self, panel_id: int) -> dict:
        """Get the raw data for a system."""
        resp = await self.__get(
            f"systems/{panel_id}",
            headers={"Accept-Encoding": "application/json"},
            params={"includerules": "false"},
        )
        if not resp:
            raise VivintSkyApiError("Unable to retrieve system data")
        return resp

    async def get_device_data(self, panel_id: int, device_id: int) -> dict:
        """Get the raw data for a device."""
        resp = await self.__get(
            f"system/{panel_id}/device/{device_id}",
            headers={"Accept-Encoding": "application/json"},
        )
        if not resp:
            raise VivintSkyApiError("Unable to retrieve device data")
        return resp

    async def set_alarm_state(
        self, panel_id: int, partition_id: int, state: int
    ) -> None:
        """Set the alarm state."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/armedstates",
            headers={"Content-Type": "application/json;charset=UTF-8"},
            data=json.dumps(
                {
                    "system": panel_id,
                    "partitionId": partition_id,
                    "armState": state,
                    "forceArm": False,
                }
            ),
        )
        if resp is None:
            _LOGGER.error(
                "Failed to set state to %s for panel %s",
                ArmedState(state).name,
                panel_id,
            )
            raise VivintSkyApiError("Failed to set alarm state")

    async def set_garage_door_state(
        self, panel_id: int, partition_id: int, device_id: int, state: int
    ) -> None:
        """Open/Close garage door."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/door/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(
                {
                    VivintDeviceAttribute.STATE: state,
                    VivintDeviceAttribute.ID: device_id,
                }
            ),
        )
        if resp is None:
            _LOGGER.debug(
                "Failed to set state to %s for garage door %s @ %s:%s",
                GarageDoorState(state).name,
                device_id,
                panel_id,
                partition_id,
            )
            raise VivintSkyApiError("Failed to set garage door state")

    async def set_lock_state(
        self, panel_id: int, partition_id: int, device_id: int, locked: bool
    ) -> None:
        """Lock/Unlock door lock."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/locks/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(
                {
                    VivintDeviceAttribute.STATE: locked,
                    VivintDeviceAttribute.ID: device_id,
                }
            ),
        )
        if resp is None:
            _LOGGER.debug(
                "Failed to set state to %s for lock %s @ %s:%s",
                locked,
                device_id,
                panel_id,
                partition_id,
            )
            raise VivintSkyApiError("Failed to set lock state")

    async def set_sensor_state(
        self, panel_id: int, partition_id: int, device_id: int, bypass: bool
    ) -> None:
        """Bypass/unbypass a sensor."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/sensors/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(
                {
                    VivintDeviceAttribute.BYPASSED: ZoneBypass.MANUALLY_BYPASSED
                    if bypass
                    else ZoneBypass.UNBYPASSED,
                    VivintDeviceAttribute.ID: device_id,
                }
            ),
        )
        if resp is None:
            _LOGGER.debug(
                "Failed to set state to %s for sensor %s @ %s:%s",
                "bypassed" if bypass else "unbypassed",
                device_id,
                panel_id,
                partition_id,
            )
            raise VivintSkyApiError("Failed to set sensor state")

    async def set_switch_state(
        self,
        panel_id: int,
        partition_id: int,
        device_id: int,
        on: bool | None = None,  # pylint: disable=invalid-name
        level: int | None = None,
    ) -> None:
        """Set switch state."""
        # validate input
        if on is None and level is None:
            raise VivintSkyApiError('Either "on" or "level" must be provided.')
        if level and (0 > level or level > 100):
            raise VivintSkyApiError('The value for "level" must be between 0 and 100.')

        data: dict = {SwitchAttribute.ID: device_id}
        if level is None:
            data[SwitchAttribute.STATE] = on
        else:
            data[SwitchAttribute.VALUE] = level

        resp = await self.__put(
            f"{panel_id}/{partition_id}/switches/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(data),
        )
        if resp is None:
            _LOGGER.debug(
                "Failed to set %s to %s for switch %s @ %s:%s",
                "on" if level is None else "level",
                on if level is None else level,
                device_id,
                panel_id,
                partition_id,
            )
            raise VivintSkyApiError("Failed to set switch state")

    async def set_thermostat_state(
        self, panel_id: int, partition_id: int, device_id: int, **kwargs: Any
    ) -> None:
        """Set thermostat state."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/thermostats/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(kwargs),
        )
        if resp is None:
            _LOGGER.debug(
                "Failed to set state to %s for thermostat %s @ %s:%s",
                kwargs,
                device_id,
                panel_id,
                partition_id,
            )
            raise VivintSkyApiError("Failed to set thermostat state")

    async def request_camera_thumbnail(
        self, panel_id: int, partition_id: int, device_id: int
    ) -> None:
        """Request the camera thumbnail."""
        try:
            await self.__get(
                f"{panel_id}/{partition_id}/{device_id}/request-camera-thumbnail",
            )
        except ClientResponseError as resp:
            if resp.status < 200 or resp.status > 299:
                _LOGGER.error(
                    "Failed to request thumbnail for camera %s @ %s:%s with status code %s",
                    device_id,
                    panel_id,
                    partition_id,
                    resp.status,
                )

    async def get_camera_thumbnail_url(
        self, panel_id: int, partition_id: int, device_id: int, thumbnail_timestamp: int
    ) -> str | None:
        """Get the camera thumbnail url."""
        try:
            resp = await self.__get(
                f"{panel_id}/{partition_id}/{device_id}/camera-thumbnail",
                params={"time": thumbnail_timestamp},
                allow_redirects=False,
            )
            assert resp
            return resp.get("location")
        except ClientResponseError as resp:
            if resp.status != 302:
                _LOGGER.debug(
                    "Failed to get thumbnail for camera %s @ %s:%s with status code %s",
                    device_id,
                    panel_id,
                    partition_id,
                    resp.status,
                )
            return None

    def __get_new_client_session(self) -> aiohttp.ClientSession:
        """Create a new aiohttp.ClientSession object."""
        ssl_context = ssl.create_default_context(
            purpose=ssl.Purpose.SERVER_AUTH, cafile=certifi.where()
        )
        connector = aiohttp.TCPConnector(enable_cleanup_closed=True, ssl=ssl_context)

        return aiohttp.ClientSession(connector=connector)

    async def __get_vivintsky_session(
        self, username: str, password: str, persist_session: bool = False
    ) -> dict:
        """Login into the Vivint Sky platform with the given username, password and, optionally, persist the session.

        Returns auth user data if successful.
        """
        resp = await self.__post(
            "login",
            data=json.dumps(
                {
                    "username": username,
                    "password": password,
                    "persist_session": persist_session,
                }
            ),
        )
        assert resp
        return resp

    async def __get(
        self,
        path: str,
        headers: dict | None = None,
        params: dict | None = None,
        allow_redirects: bool | None = None,
    ) -> dict | None:
        """Perform a get request."""
        return await self.__call(
            self.__client_session.get,
            path,
            headers=headers,
            params=params,
            allow_redirects=allow_redirects,
        )

    async def __post(self, path: str, data: str | None = None) -> dict | None:
        """Perform a post request."""
        return await self.__call(self.__client_session.post, path, data=data)

    async def __put(
        self, path: str, headers: dict | None = None, data: str | None = None
    ) -> dict | None:
        """Perform a put request."""
        return await self.__call(
            self.__client_session.put, path, headers=headers, data=data
        )

    async def __call(
        self,
        method: Callable[..., _RequestContextManager],
        path: str,
        headers: dict | None = None,
        params: dict | None = None,
        data: str | None = None,
        allow_redirects: bool | None = None,
    ) -> dict | None:
        """Perform a request with supplied parameters and reauthenticate if necessary."""
        if path != "login" and not self.is_session_valid():
            await self.connect()

        if self.__client_session.closed:
            raise VivintSkyApiError("The client session has been closed")

        is_mfa_request = path == VIVINT_MFA_ENDPOINT

        if self.__mfa_pending and not is_mfa_request:
            raise VivintSkyApiMfaRequiredError(AuthenticationResponse.MFA_REQUIRED)

        resp = await method(
            path if is_mfa_request else f"{VIVINT_API_ENDPOINT}/{path}",
            headers=headers,
            params=params,
            data=data,
            allow_redirects=allow_redirects,
        )
        async with resp:
            resp_data: dict = await resp.json(encoding="utf-8")
            if resp.status == 200:
                return resp_data
            if resp.status == 302:
                return {"location": resp.headers.get("Location")}
            if resp.status in (400, 401, 403):
                message = (
                    resp_data.get(MfaVerificationResponse.MESSAGE)
                    if is_mfa_request
                    else resp_data.get(AuthenticationResponse.MESSAGE)
                )
                if message == AuthenticationResponse.MFA_REQUIRED or is_mfa_request:
                    self.__mfa_pending = True
                    raise VivintSkyApiMfaRequiredError(message)
                raise VivintSkyApiAuthenticationError(message)
            resp.raise_for_status()
            return None



"""Module that defines various enums."""
from __future__ import annotations

import logging
from enum import Enum, IntEnum, unique
from typing import Any

_LOGGER = logging.getLogger(__name__)


@unique
class ArmedState(IntEnum):
    """Armed state."""

    DISARMED = 0
    ARMING_AWAY_IN_EXIT_DELAY = 1
    ARMING_STAY_IN_EXIT_DELAY = 2
    ARMED_STAY = 3
    ARMED_AWAY = 4
    ARMED_STAY_IN_ENTRY_DELAY = 5
    ARMED_AWAY_IN_ENTRY_DELAY = 6
    ALARM = 7
    ALARM_FIRE = 8
    DISABLED = 11
    WALK_TEST = 12

    # Handle unknown/future armed state
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: object) -> ArmedState:
        _LOGGER.debug("Unknown armed state value: %s", value)
        return cls.UNKNOWN


@unique
class CapabilityCategoryType(IntEnum):
    """Capability category type."""

    CAMERA = 1
    DOORBELL = 2
    SWITCH = 3
    THERMOSTAT = 4
    ZWAVE = 5
    TOUCHLINK = 6
    DOOR_LOCK = 7
    MOBILE_BLACKLIST = 8
    GARAGE_DOOR = 9
    KEYPAD = 10
    PLATFORM = 11

    # Handle unknown/future capability category type
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: object) -> CapabilityCategoryType:
        _LOGGER.debug("Unknown capability category type value: %s", value)
        return cls.UNKNOWN


class CapabilityType(IntEnum):
    """Capability type."""

    ANIMAL_DETECTION = 42
    BINARY_ON_OFF = 7
    CAMERA_DVR_CAPABLE = 72
    CAN_CHIME = 4
    CAN_ENABLE_DISABLE_AUDIO = 73
    CAN_ENABLE_DISABLE_LED = 74
    CAN_INITIATE_CALL = 66
    CAN_LOCK_UNLOCK = 30
    CANNOT_REPORT_STATUS = 33
    CHIME_EXTENDER = 36
    CLIP_CAPTURE = 3
    DELETE_ALL_EVENTS = 63
    DETER = 45
    DETER_LIGHT = 52
    DETER_LINGER_DURATION = 53
    DETER_SCHEDULE = 50
    DETER_TONE = 51
    DETER_ZONE = 49
    DIMMABLE = 8
    DOOR_STATE = 71
    DOORBELL_CHIME_SELECTABLE = 56
    FAN120_MINUTE = 23
    FAN15_MINUTE = 18
    FAN240_MINUTE = 24
    FAN30_MINUTE = 19
    FAN45_MINUTE = 20
    FAN480_MINUTE = 25
    FAN60_MINUTE = 21
    FAN960_MINUTE = 26
    HAS_MICROPHONE = 46
    HAS_USER_CODE = 75
    HSB_LIGHTING = 9
    HUMIDITY = 64
    HUMIDITY_CONTROL = 65
    IN_HOME_CHIME_VOLUME = 54
    LED_BEEPER_SETTINGS = 73
    LIVE_VIDEO = 6
    LURKER_DETECTION = 38
    MAINTAIN_ZOOM = 48
    MAX_COOL_LOCK_SET_POINT = 69
    MAX_HEAT_LOCK_SET_POINT = 67
    MAXIMUM_TEMPERATURE = 12
    MIN_COOL_LOCK_SET_POINT = 70
    MIN_HEAT_LOCK_SET_POINT = 68
    MIN_SETPOINT_DIFFERENCE_C = 13
    MIN_SETPOINT_DIFFERENCE_F = 14
    MINIMUM_TEMPERATURE = 11
    MOBILE_BLACKLISTABLE = 31
    MOTION_DETECTION = 1
    MUTE_CHIME = 55
    NIGHT_VISION = 60
    PACKAGE_DETECTION = 39
    PACKAGE_MOVE_DETECTION = 40
    PERSON_DETECTION = 16
    PINCH_TO_ZOOM = 32
    POLYGON_ROI = 44
    PREVIEW_CHIME_IN_HOME = 58
    PRIVACY_MODE = 15
    QUIET_MODE = 17
    REBOOT_CAMERA = 62
    RECT_ROI = 47
    RESTORE_DEFAULTS = 61
    ROTATE_IMAGE = 34
    SECURITY_2_0 = 27
    SET_POINT_LOCK = 66
    SIREN_EXTENDER = 37
    SOFT_DIP_SWITCH = 29
    STATUS_LIGHT_TOGGLE = 35
    TEMPERATURE_DETECTION = 10
    TWO_WAY_AUDIO = 5
    VEHICLE_DETECTION = 41
    VIDEO_QUALITY = 59
    VISITOR_CHIME_SELECTABLE = 57
    VISITOR_DETECTION = 2
    WARPED = 43
    WI_FI = 28
    ZWAVE_DIAGNOSTICS = 22

    # Handle unknown/future capability type
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: Any) -> CapabilityType:
        _LOGGER.debug("Unknown capability type value: %s", value)
        return cls.UNKNOWN


# @unique
class DeviceType(Enum):
    """Device type."""

    AIR_TOWER = "airtower_device"
    ALPHA_CS6022_CAMERA_DEVICE = "alpha_cs6022_camera_device"
    BINARY_SWITCH = "binary_switch_device"
    CAMERA = "camera_device"
    CARGUARD_DEVICE = "carguard_device"
    DISPLAY_DEVICE = "display_device"
    DOOR_LOCK = "door_lock_device"
    ENERGY_SERVICE = "energy_service"
    GARAGE_DOOR = "garage_door_device"
    GROUP_DEVICE = "group_device"
    HSB_SWITCH = "multilevel_switch_ip_device"
    HUE_BRIDGE = "phillips_hue_bridge_device"
    IOT_SERVICE = "iot_service"
    KEY_FOB = "keyfob_device"
    KEY_PAD = "keypad_device"
    KWIKSET988_DOOR_LOCK = "kwikset_988_door_lock_device"
    MULTI_LEVEL_SWITCH = "multilevel_switch_device"
    NEST_THERMOSTAT = "nest_thermostat_device"
    NETWORK_HOSTS_SERVICE = "network_hosts_service"
    PANEL = "primary_touch_link_device"
    PANEL_DIAGNOSTICS_SERVICE = "panel_diagnostics_service"
    PERSONAL_GUARD_DEVICE = "personal_guard_device"
    POD_MY_Q_GARAGE_DOOR_DEVICE = "pod_myq_garage_door_device"
    POD_NEST_THERMOSTAT = "pod_nest_thermostat_device"
    SCHEDULER_SERVICE = "scheduler_service"
    SECURITY_SERVICE = "security_service"
    SENSOR_GROUP = "sensor_group"
    SLIMLINE = "slim_line_device"
    SMART_THERMOSTAT = "ct200_thermostat_device"
    SMART_THERMOSTAT_V2 = "ev2_thermostat_device"
    SPACE_MONKEY = "space_monkey_service"
    THERMOSTAT = "thermostat_device"
    VIVOTEK620_PT_CAMERA = "vivotek_620pt_camera_device"
    VIVOTEK720_CAMERA = "vivotek_720_camera_device"
    VIVOTEK720_W_CAMERA = "vivotek_720w_camera_device"
    VIVOTEK_DB8331_W_CAMERA = "vivotek_db8331w_camera_device"
    VIVOTEK_DB8332_CAMERA = "vivotek_db8332_camera_device"
    VIVOTEK_DB8332_S1_CAMERA = "vivotek_db8332s1_camera_device"
    VIVOTEK_DB8332_SW_CAMERA = "vivotek_db8332sw_camera_device"
    VIVOTEK_HD400_W_CAMERA = "vivotek_hd400w_camera_device"
    VIVOTEK_HDP450_CAMERA = "vivotek_hdp450_camera_device"
    WIRED_SENSOR = "wired_sensor"
    WIRELESS_SENSOR = "wireless_sensor"
    YOFI_DEVICE = "yofi_device"
    ZWAVE_KEY_PAD = "keypad_entry_device"

    # Deprecated
    MULTILEVEL_SWITCH = MULTI_LEVEL_SWITCH
    TOUCH_PANEL = PANEL

    UNKNOWN = None

    @classmethod
    def _missing_(cls, value: Any) -> DeviceType:
        _LOGGER.debug("Unknown device type value: %s", value)
        return cls.UNKNOWN


@unique
class EquipmentCode(IntEnum):
    """Equipment code."""

    APOLLO_COMBO_CO = 4050
    APOLLO_COMBO_SMOKE = 4040
    CARBON_MONOXIDE_DETECTOR_345_MHZ = 1254
    CO1_CO = 860
    CO1_CO_CANADA = 859
    CO3_2_GIG_CO = 1026
    DBELL1_2_GIG_DOORBELL = 1063
    DW10_THIN_DOOR_WINDOW = 862
    DW11_THIN_DOOR_WINDOW = 1251
    DW12_THIN_DOOR_WINDOW = 4000
    DW20_RECESSED_DOOR = 863
    DW21_R_RECESSED_DOOR = 1252
    EXISTING_CO = 692
    EXISTING_DOOR_WINDOW_CONTACT = 655
    EXISTING_FLOOD_TEMP = 556
    EXISTING_GLASS_BREAK = 475
    EXISTING_HEAT = 708
    EXISTING_KEY_FOB_REMOTE = 577
    EXISTING_MOTION_DETECTOR = 609
    EXISTING_SMOKE = 616
    FIREFIGHTER_AUDIO_DETECTOR = 1269
    GARAGE01_RESOLUTION_TILT = 1061
    GB1_GLASS_BREAK = 864
    GB2_GLASS_BREAK = 1248
    GB3_GLASS_BREAK = 4030
    HW_DW_5816 = 673
    HW_FLOOD_SENSOR_5821 = 624
    HW_GLASS_BREAK_5853 = 519
    HW_HEAT_SENSOR_5809 = 557
    HW_PANIC_PENDANT_5802_MN2 = 491
    HW_PIR_5890 = 533
    HW_PIR_5894_PI = 530
    HW_R_DW_5818_MNL = 470
    HW_SMOKE_5808_W3 = 589
    KEY1_345_4_BUTTON_KEY_FOB_REMOTE = 866
    OTHER = 0
    PAD1_345_WIRELESS_KEYPAD = 867
    PANIC1 = 868
    PANIC2 = 1253
    PANIC3 = 4130
    PIR1_MOTION = 869
    PIR2_MOTION = 1249
    PIR3_MOTION = 4020
    RE219_FLOOD_SENSOR = 1128
    RE220_T_2_GIG_REPEATER = 1144
    RE224_DT_DSC_TRANSLATOR = 1208
    RE224_GT_GE_TRANSLATOR = 941
    RE508_X_REPEATER = 2832
    RE524_X_WIRELESS_TAKEOVER = 2830
    REPEATER_345_MHZ = 2081
    SECURE_KEY_345_MHZ = 1250
    SMKE1_SMOKE = 872
    SMKE1_SMOKE_CANADA = 871
    SMKT2_GE_SMOKE_HEAT = 895
    SMKT3_2_GIG = 1058
    SMKT6_2_GIG = 1066
    SWS1_SMART_WATER_SENSOR = 1264
    TAKE_TAKEOVER = 873
    TILT_SENSOR_2_GIG_345 = 2831
    VS_CO3_DETECTOR = 1266
    VS_SMKT_SMOKE_DETECTOR = 1267

    # Handle unknown/future equipment code
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: Any) -> EquipmentCode:
        _LOGGER.debug("Unknown equipment code value: %s", value)
        return cls.UNKNOWN


@unique
class EquipmentType(IntEnum):
    """Equipment type."""

    CONTACT = 1
    EMERGENCY = 11
    FREEZE = 6
    MOTION = 2
    TEMPERATURE = 10
    WATER = 8

    # Handle unknown/future equipment type
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: Any) -> EquipmentType:
        _LOGGER.debug("Unknown equipment type value: %s", value)
        return cls.UNKNOWN


@unique
class FanMode(IntEnum):
    """Fan mode."""

    AUTO_LOW = 0
    ON_LOW = 1
    AUTO_HIGH = 2
    ON_HIGH = 3
    TIMER_15 = 99
    TIMER_30 = 100
    TIMER_45 = 102
    TIMER_60 = 101
    TIMER_120 = 103
    TIMER_240 = 104
    TIMER_480 = 105
    TIMER_720 = 107
    TIMER_960 = 106

    # Handle unknown/future fan mode
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: Any) -> FanMode:
        _LOGGER.debug("Unknown fan mode value: %s", value)
        return cls.UNKNOWN


@unique
class GarageDoorState(IntEnum):
    """Garage door state."""

    UNKNOWN = 0
    CLOSED = 1
    CLOSING = 2
    STOPPED = 3
    OPENING = 4
    OPENED = 5

    @classmethod
    def _missing_(cls, value: Any) -> GarageDoorState:
        _LOGGER.debug("Unknown garage door state value: %s", value)
        return cls.UNKNOWN


@unique
class HoldMode(IntEnum):
    """Hold mode."""

    BY_SCHEDULE = 0
    UNTIL_NEXT = 1
    TWO_HOURS = 2
    PERMANENT = 3

    # Handle unknown/future hold mode
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: Any) -> HoldMode:
        _LOGGER.debug("Unknown hold mode value: %s", value)
        return cls.UNKNOWN


@unique
class OperatingMode(IntEnum):
    """Operating mode."""

    OFF = 0
    HEAT = 1
    COOL = 2
    AUTO = 3
    EMERGENCY_HEAT = 4
    RESUME = 5
    FAN_ONLY = 6
    FURNACE = 7
    DRY_AIR = 8
    MOIST_AIR = 9
    AUTO_CHANGEOVER = 10
    ENERGY_SAVE_HEAT = 11
    ENERGY_SAVE_COOL = 12
    AWAY = 13
    ECO = 100

    # Handle unknown/future operating mode
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: Any) -> OperatingMode:
        _LOGGER.debug("Unknown operating mode value: %s", value)
        return cls.UNKNOWN


@unique
class OperatingState(IntEnum):
    """Operating state."""

    IDLE = 0
    HEATING = 1
    COOLING = 2

    # Handle unknown/future operating state
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: Any) -> OperatingState:
        _LOGGER.debug("Unknown operating state value: %s", value)
        return cls.UNKNOWN


@unique
class SensorType(IntEnum):
    """Sensor type."""

    AUDIBLE_ALARM = 7
    AUXILIARY_ALARM = 8
    CARBON_MONOXIDE = 14
    DAY_ZONE = 5
    EXIT_ENTRY_1 = 1
    EXIT_ENTRY_2 = 2
    FIRE = 9
    FIRE_WITH_VERIFICATION = 16
    INTERIOR_FOLLOWER = 4
    INTERIOR_WITH_DELAY = 10
    NO_RESPONSE = 23
    PERIMETER = 3
    REPEATER = 25
    SILENT_ALARM = 6
    SILENT_BURGLARY = 24
    UNUSED = 0

    # Handle unknown/future sensor type.
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: Any) -> SensorType:
        _LOGGER.debug("Unknown sensor type value: %s", value)
        return cls.UNKNOWN


@unique
class ZoneBypass(IntEnum):
    """Zone bypass statuses."""

    UNBYPASSED = 0
    FORCE_BYPASSED = 1
    MANUALLY_BYPASSED = 2

    # Handle unknown/future zone bypass.
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: Any) -> ZoneBypass:
        _LOGGER.debug("Unknown zone bypass value: %s", value)
        return cls.UNKNOWN
